-- SSI Design Laboratory
-- Digital implementation flow
-- testbench.vhd
-- Péter Horváth, 2014




----------------------------------------------------------------------------------------------------
-- This file includes a high-level model of the frequency measurement circuitry augmented with an
-- UART transmitter and an SPI slave module, which are used to transmit the measurement data
-- to the outside world.
-- 
-- Take into consideration that this model is NOT SYNTHESIZABLE.

-- The frequency measurement circuit works as follows:
-- The circuit counts the edges of the input signal in a well-defined measurement
-- window, which is determined by the parameter measurement_window. 
-- If we know the number of rising edges of a signal occured
-- in a well-known time interval, the frequency can be calculated:
-- frequency = 1 / (T_window / edge_number), where
-- T_window: time interval in which the edges are counted
-- edge_number: the number of positive edges counted during the time interval
-- The frequency value calculated this way is an average frequency related to
-- the recent timw window.

-- Since the division is a very complicated and expensive operation for hardware
-- implementation, only the high-level behavioral model performs this operation. 
-- The cycle-accurate, synthesizable model should not calculate the frequency, 
-- it only has to determine the number of edges counted during the measurement window
-- and to send this data through UART or SPI.
----------------------------------------------------------------------------------------------------





library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
----------------------------------------------------------------------------------------------------
entity FMEAS is
	port (
		-------------------------------------------------------
		-- parameters of the frequency measurement circuitry --
		-------------------------------------------------------
		measure_window:				in	time := 0.5 ms;							-- frequency measurement time window
		comm_protocol:				in	integer range 0 to 1 := 0;				-- 0: UART; 1: SPI
		bytes_per_frame:			in	integer range 2 to 3 := 3;				-- the number of bytes the counter valuies are stored into
		UART_bps:					in	integer range 150 to 256000 := 256000;	-- UART baud rate
		UART_stop_bits:				in	integer range 1 to 2 := 2;				-- UART format: 8N1 or 8N2
		SPI_CPOL:					in	bit := '0';								-- SPI clock polarity
		SPI_CPHA:					in	bit := '0';								-- SPI clock phase
		
		--------------------------------------------------
		-- ports of the frequency measurement circuitry --
		--------------------------------------------------
		reset:						in	std_logic;								-- global reset signal
		digitalized_value:			in	std_logic;								-- the digitalized value of the input signal, whose frequency should be measured
		TX:							out	std_logic := 'Z';						-- UART TX line driven by FMEAS
		SCLK:						in	std_logic;								-- SPI SCLK signal driven by the SPI master of the testbench
		SS:							in	std_logic;								-- SPI SS signal driven by the SPI master of the testbench
		MISO:						out	std_logic := 'Z'						-- SPI MISO signal driven by FMEAS
	);
end FMEAS;
----------------------------------------------------------------------------------------------------
architecture behavior of FMEAS is

	signal reset_edge_counter: 			std_logic := '0';
	signal edge_counter:				integer range 0 to 2**24-1 := 0;
	signal new_data:					integer range 0 to 2**24-1 := 0;
	signal UART_send_trigger:			std_logic := '0';
	signal measured_cycle_time:			time := 1 ns;
	signal measured_frequency_MHz:		real;

begin
	-----------------------------------------------------------------
	-- MEASURING THE INPUT FREQUENCY --
	-- The process L_MEASURE_WINDOW generates a clock-like signal,
	-- whose rising edges indicate the start of a new measurement
	-- window.
	-----------------------------------------------------------------
	
	
	L_MEASURE_WINDOW: process
	begin
		wait for measure_window / 2;
		reset_edge_counter <= not reset_edge_counter;
	end process;

	L_COUNT_EDGES: process ( reset, reset_edge_counter, digitalized_value )
	begin
		if ( reset = '1' ) then
			edge_counter <= 0;
			new_data <= 0;
			measured_cycle_time <= 1 ns;
			UART_send_trigger <= '0';
		elsif ( rising_edge(digitalized_value) ) then
			edge_counter <= edge_counter + 1;
		elsif ( rising_edge(reset_edge_counter) ) then
			new_data <= edge_counter;
			measured_cycle_time <= measure_window / real(edge_counter);
			edge_counter <= 0;
			UART_send_trigger <= not UART_send_trigger;
		end if;
	end process;
	
	measured_frequency_MHz <= 1.0 / ( real(measured_cycle_time / 1 ps) / 1000000.0);
	
	L_WRITE_FREQUENCIES: process ( measured_frequency_MHz )
		file frequencies_Hz: text open write_mode is "digit_output_without_header.csv";
		variable line_buffer: line;
	begin
		if ( now > 4*measure_window ) then
			write(line_buffer, real(now / 1 ps) * 10.0**(-12) );
			write(line_buffer, string'(","));
			write(line_buffer, measured_frequency_MHz * 10.0**6);
			writeline(frequencies_Hz, line_buffer);
		end if;
	end process;
	
	
	-----------------------------------------------------------------
	-- GENERATING UART BITSTREAM --
	-----------------------------------------------------------------
	
	
	L_UART_TRANSMITTER: process
	
		procedure UART_send(integer_2_send: in integer; signal TX: out std_logic) is
		begin
			TX <= '0';
			wait for (1.0/real(UART_bps)) * 1 sec;
			for i in 0 to 7 loop
				TX <= to_unsigned(integer_2_send, 8)(i);
				wait for (1.0/real(UART_bps)) * 1 sec;
			end loop;
			TX <= '1';
			if ( UART_stop_bits = 2 ) then wait for (2.0/real(UART_bps)) * 1 sec;
			else wait for (1.0/real(UART_bps)) * 1 sec;
			end if;
		end procedure;
		
		variable integer_2_send: integer := 0;
	
	begin
		wait on UART_send_trigger;
		for i in 1 to bytes_per_frame loop
			integer_2_send := to_integer(to_unsigned(new_data, 8*bytes_per_frame)((i*8)-1 downto (i*8)-8));
			if ( comm_protocol = 0 ) then UART_send(integer_2_send, TX); end if;
		end loop;
	end process;
	
	
	-----------------------------------------------------------------
	-- SPI SLAVE INTERFACE --
	-----------------------------------------------------------------
	
	
	L_SPI_SLAVE: process
		variable frame_2_send: unsigned (23 downto 0) := (others => '0');
	begin
		wait until falling_edge(SS);
		frame_2_send := to_unsigned(new_data, 24);
		if ( SPI_CPHA = '0' ) then 
			
			if ( comm_protocol = 1 ) then MISO <= frame_2_send(0); end if;
			
			for i in 1 to 8*bytes_per_frame-1 loop
				if ( SPI_CPOL = '0' ) then wait until falling_edge(SCLK);
				else wait until rising_edge(SCLK);
				end if;
				
				if ( comm_protocol = 1 ) then MISO <= frame_2_send(i); end if;
			
			end loop;
		else
			for i in 0 to 8*bytes_per_frame-1 loop
				if ( SPI_CPOL = '0' ) then wait until rising_edge(SCLK);
				else wait until falling_edge(SCLK);
				end if;
				
				if ( comm_protocol = 1 ) then MISO <= frame_2_send(i); end if;
		
			end loop;
		end if;
	end process;
	
	
	-----------------------------------------------------------------
	-----------------------------------------------------------------
	-----------------------------------------------------------------
end behavior;
----------------------------------------------------------------------------------------------------



----------------------------------------------------------------------------------------------------
-- This file includes a VHDL test environment, whose objective is two-fold:
--			-	it reads the transient response of the VCO and performs a simple
--				comparator-based analog to digital conversion
--			- 	it instantiates the frequency measurement circuit and provides it
--				with input stimuli including a reset signal and an SPI master
--				module, which generates read requests for the SPI slave interface
--				of the frequency measurement circuit
----------------------------------------------------------------------------------------------------




library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use std.textio.all;
----------------------------------------------------------------------------------------------------
entity testbench is
end testbench;
----------------------------------------------------------------------------------------------------
architecture behavior of testbench is

	------------------------------------------------
	-- input signals of the measurement circuitry --
	------------------------------------------------
	signal reset:								std_logic := '0';	-- global reset signal
	signal SCLK:								std_logic := '0'; 	-- SPI SCLK signal generated by the SPI master
	signal SS:									std_logic := '1'; 	-- SPI SS signal generated by the SPI master
	
	
	
	-------------------------------------------------
	-- output signals of the measurement circuitry --
	-------------------------------------------------
	signal TX:					  				std_logic;			-- UART TX signal generated by the UART transmitter of the frequency measurement circuit (process L_UART_TRANSMITTER in architecture behavior of entity FMEAS)
	signal MISO:								std_logic;			-- SPI MISO signal generated by the SPI slave module of the frequency measurement circuit (process L_SPI_SLAVE in architecture behavior of entity FMEAS)
	
	
	
	----------------------------------
	-- additional testbench signals --
	----------------------------------
	signal measure_window:								time := 0.5 ms;
	signal comm_protocol:								integer range 0 to 1 := 0;			
	signal bytes_per_frame:								integer range 2 to 3 := 3;
	signal UART_bps:									integer range 150 to 256000 := 256000;
	signal UART_stop_bits:								integer range 1 to 2 := 2;
	signal SPI_CPOL:									bit := '0';			
	signal SPI_CPHA:									bit := '0';			
	signal sim_time:									time := 100 ms;
	
	signal end_of_file:									std_logic := '0';
	signal analog_value: 								real := 0.0;
	signal digitalized_value:							std_logic;
	signal frame_read_by_SPI_master_tmp:				std_logic_vector (23 downto 0) := (others => '0');
	signal frame_read_by_SPI_master:					std_logic_vector (23 downto 0) := (others => '0');
	signal UART_received_data_tmp:						std_logic_vector (7 downto 0) := (others => '0');
	signal UART_received_data:							std_logic_vector (7 downto 0) := (others => '0');


begin

	----------------------------------
	-- Reading parameters from file --
	-- 		v_measure_window:		the time interval, in which the edges of digitalized input signal (digitalized_value) are counted
	--		v_comm_protocol:		0: UART, 1: SPI. The unselected communication module is inactive, which means that its serial output is Z
	--		v_bytes_per_frame:		2 or 3. The number of bytes representing a single counter value
	--		v_UART_bps:				baud rate of the UART communication
	--		v_UART_stop_bits:		1 or 2. UART format: 8N1 or 8N2
	--		V_SPI_CPOL:				SCLK polarity of the SPI communication
	--		v_SPI_CPHA:				SCLK phase of the SPI communication
	----------------------------------
	L_READ_GENERICS: process
	
		file parameters: text open read_mode is "parameters.conf";
		variable line_buffer_parameters: 			line;
		variable v_measure_window:					real := 0.1e-3;
		variable v_comm_protocol:					integer range 0 to 1 := 0;
		variable v_bytes_per_frame:					integer range 2 to 3 := 3;
		variable v_UART_bps:						integer range 150 to 256000 := 256000;
		variable v_UART_stop_bits:					integer range 1 to 2;
		variable v_SPI_CPOL:						bit := '0';
		variable v_SPI_CPHA:						bit := '0';
	
	begin
		readline(parameters, line_buffer_parameters);
		read(line_buffer_parameters, v_measure_window);
		
		readline(parameters, line_buffer_parameters);
		read(line_buffer_parameters, v_UART_bps);
		
		readline(parameters, line_buffer_parameters);
		read(line_buffer_parameters, v_UART_stop_bits);
		
		readline(parameters, line_buffer_parameters);
		read(line_buffer_parameters, v_SPI_CPOL);
		
		readline(parameters, line_buffer_parameters);
		read(line_buffer_parameters, v_SPI_CPHA);
		
		readline(parameters, line_buffer_parameters);
		read(line_buffer_parameters, v_bytes_per_frame);
		
		readline(parameters, line_buffer_parameters);
		read(line_buffer_parameters, v_comm_protocol);
		
		measure_window <= (v_measure_window * 1 sec);
		comm_protocol <= v_comm_protocol;
		bytes_per_frame <= v_bytes_per_frame;
		UART_bps <= v_UART_bps;
		UART_stop_bits <= v_UART_stop_bits;
		SPI_CPOL <= v_SPI_CPOL;
		SPI_CPHA <= v_SPI_CPHA;
		
		wait;
	end process;
	
	
	
	----------------------------------------------------------------------------------------------
	-- This process reads the transient response generated by the analog simulator (transient.csv).
	-- The analog value is represented by the signal called analog_value.
	----------------------------------------------------------------------------------------------
	L_GENERATE_ANALOG_SIGNAL: process
		file transient: text open read_mode is "../results/analog_output.csv";
		variable line_buffer_transient: line;
		variable previous_time: time := 0 ps;
		variable recent_time_real_sec: real := 0.0;
		variable recent_time: time := 0 ps;
		variable recent_value: real := 0.0;
		variable line_number: integer := 0;
	begin
		if ( not endfile(transient) ) then
			previous_time := recent_time;
			readline(transient, line_buffer_transient);
			line_number := line_number + 1;
			read(line_buffer_transient, recent_time_real_sec);
			read(line_buffer_transient, recent_value);
			recent_time := recent_time_real_sec * 1 sec;
			wait for (recent_time - previous_time);
			analog_value <= recent_value;
		else
			end_of_file <= '1';
			file_close(transient);
			wait;
		end if;
	end process;
	
	
	
	------------------------------------------------------------------------------
	-- The A/D converter is a comparator. If the analog value is greater than 1.65,
	-- the digitalized signal (digitalized_value) 
	-- gets logic one otherwise it gets logic zero.
	------------------------------------------------------------------------------
	L_AD: process ( analog_value )
	begin
		if ( analog_value > 1.65 ) then digitalized_value <= '1';
		else digitalized_value <= '0';
		end if;
	end process;
							
							
							
	----------------------------------------------------------------------------------------
	-- Instantiation of the frequency measurement circuit's behavioral model called FMEAS --
	----------------------------------------------------------------------------------------
	L_FMEAS: 	entity work.FMEAS(behavior)
							port map (	-------------------------------------------------------
										-- parameters of the frequency measurement circuitry --
										-------------------------------------------------------
										measure_window => measure_window,
										comm_protocol => comm_protocol,
										bytes_per_frame => bytes_per_frame,
										UART_bps => UART_bps,
										UART_stop_bits => UART_stop_bits,
										SPI_CPOL => SPI_CPOL,
										SPI_CPHA => SPI_CPHA,
										
										--------------------------------------------------
										-- ports of the frequency measurement circuitry --
										--------------------------------------------------
										reset => reset,
										digitalized_value	=> digitalized_value,
										TX => TX,
										SCLK => SCLK,
										SS => SS,
										MISO => MISO);


	
	-------------------------------------------------------------------------------------------
	-- The frequency measurement circuitry may have an SPI slave interface.                  --
	-- This process models an SPI master, which generates a read request for the SPI slave.  --
	-- The delay between two read requests is twice as long as the measrement window itself. --
	-- The SCLK frequency of the SPI master is 1 MHz.																				 --
	-------------------------------------------------------------------------------------------
	L_SPI_MASTER: process
	begin
		wait for 1 ps;
		if ( comm_protocol = 1 ) then
			wait for 2*measure_window;
			if ( SPI_CPOL = '1' ) then SCLK <= '1'; else SCLK <= '0'; end if;
			SS <= '0';
			wait for 0.5 us;
			for i in 0 to 8*bytes_per_frame-1 loop
				if ( SPI_CPOL = '0' ) then 
					SCLK <= '1';
					if ( SPI_CPHA = '0' ) then frame_read_by_SPI_master_tmp(i) <= MISO; end if;
				else
					SCLK <= '0';
					if ( SPI_CPHA = '0' ) then frame_read_by_SPI_master_tmp(i) <= MISO; end if;
				end if;
				wait for 0.5 us;
				if ( SPI_CPOL = '0' ) then
					SCLK <= '0';
					if ( SPI_CPHA = '1' ) then frame_read_by_SPI_master_tmp(i) <= MISO; end if;
				else 
					SCLK <= '1';
					if ( SPI_CPHA = '1' ) then frame_read_by_SPI_master_tmp(i) <= MISO; end if;
				end if;
				wait for 0.5 us;
			end loop;
			frame_read_by_SPI_master <= frame_read_by_SPI_master_tmp;
			SS <= '1';
		else
			wait;
		end if;
	end process;
	
	
	-----------------------------------------------------------------
	-- UART RECEIVER --
	-----------------------------------------------------------------
	
	
	L_UART_RECEIVER: process
	begin
		wait for 1 ps;
		if ( comm_protocol = 0 ) then
			wait until TX'event and TX = '0';
			wait for (1.5/real(UART_bps)) * 1 sec;
			for i in 0 to 7 loop
				UART_received_data_tmp(i) <= TX;
				wait for (1.0/real(UART_bps)) * 1 sec;
			end loop;
			UART_received_data <= UART_received_data_tmp;
		else
			wait;
		end if;
	end process;
	
	
	
	--------------------------
	-- generate reset pulse --
	--------------------------
	L_RESET: process
	begin
		wait for 430 ns;
		reset <= '1';
		wait for 523 ns;
		reset <= '0';
		wait;
	end process;

end behavior;
----------------------------------------------------------------------------------------------------
